""""\
Usage: falderal [<option>...] <filename.markdown>...
"""

from optparse import OptionParser
import sys

from falderal.objects import Document, FalderalSyntaxError


##### Exceptions #####

class FalderalLintingError(ValueError):
    pass


##### Main #####

def main(args):
    parser = OptionParser()
    parser.add_option("-b", "--substring-error",
                      action="store_true", default=False,
                      help="match expected errors as substrings")
    parser.add_option("--cavalier",
                      action="store_true", default=False,
                      help="don't perform sanity linting before running tests")
    parser.add_option("-d", "--dump",
                      action="store_true", default=False,
                      help="print out info about parsed tests, don't run them")
    parser.add_option("-t", "--test",
                      action="store_true", default=False,
                      help="run internal tests only and exit")
    parser.add_option("-v", "--verbose",
                      action="store_true", default=False,
                      help="print out info about each test as it is run")

    (options, args) = parser.parse_args(args[1:])

    if options.test:
        import doctest
        import falderal.objects
        (failure_count, test_count) = \
            doctest.testmod(falderal.objects,
                            optionflags=doctest.NORMALIZE_WHITESPACE)
        if failure_count > 0:
            return 1
        else:
            return 0

    # load Documents and create Falderal Tests from them
    documents = []
    functionalities = {}
    tests = []
    try:
        for filename in args:
            documents.append(Document.load(filename))
        for document in documents:
            tests += document.extract_tests(functionalities)

        if not options.cavalier:
            if not documents:
                raise FalderalLintingError("No test documents were specified")
            if not functionalities:
                raise FalderalLintingError(
                    "No functionalities were found in any of the test documents"
                )
            for name in functionalities:
                functionality = functionalities[name]
                if not functionality.implementations:
                    raise FalderalLintingError(
                        "No implementations were found for the functionality '%s'" %
                        functionality.name
                    )
            if not tests:
                raise FalderalLintingError(
                    "No tests were found in any of the test documents"
                )

    except (FalderalSyntaxError, FalderalLintingError) as e:
        # if options.show_full_exception: do that, else
        sys.stderr.write('%s: %s\n' % (e.__class__.__name__, str(e)))
        return 1

    if options.dump:
        print "Functionalities:"
        for name in functionalities:
            print "  " + name
            for implementation in functionalities[name].implementations:
                print "  +-" + str(implementation)
        print "Tests:"
        for test in tests:
            print "  " + str(test)
        return 0

    # run tests
    results = []
    dup_check = {}
    all_ran = False
    try:
        for test in tests:
            if options.verbose:
                print str(test)
            these_results = test.run(options=options)
            if options.verbose:
                for result in these_results:
                    key = repr(test.body_block.text()) + repr(result.implementation)
                    location = "%s, line %d" % (
                        test.body_block.filename, test.body_block.line_num
                    )
                    dup_check.setdefault(key, []).append(location)
            results += these_results
        all_ran = True
    except KeyboardInterrupt:
        pass

    if options.verbose:
        for key in dup_check:
            if len(dup_check[key]) != 1:
                print "WARNING: test/impl combination %s was run %d times %r" % (
                    key, len(dup_check[key]), dup_check[key]
                )

    # report on results
    for result in results:
        result.report()
    num_results = len(results)
    num_failures = len([x for x in results if not x.is_successful()])
    if not all_ran:
        print '**************************************************************'
        print '** TESTING TERMINATED PREMATURELY -- NOT ALL TESTS WERE RUN **'
        print '**************************************************************'

    print '--------------------------------'
    print 'Total test runs: %d, failures: %d' % (num_results, num_failures)
    print '--------------------------------'

    if num_failures == 0:
        return 0
    else:
        return 1
